/* -----------------------------------------------------------------------------
 *
 *
 *  ____                   _____     _
 * / ___|  __ _ _ __  _ __|__  /  __| | ___
 * \___ \ / _` | '_ \| '_ \ / /  / _` |/ _ \
 *  ___) | (_| | |_) | |_) / /_ | (_| |  __/
 * |____/ \__,_| .__/| .__/____(_)__,_|\___|
 *             |_|   |_|
 *
 *
 *
 * Author: Andreas Gschossmann
 * Email:  andreas.gschossmann@hs-regensburg.de
 *
 * quadrotor_dynamics.cc - Relationship between rotor-speed or alternatively 
 * digital value for brushless drivers and resulting thrust of the motor.
 *
 * -----------------------------------------------------------------------------*/
#include <cmath>
#include "quadrotor_dynamics.hh"

//using namespace std;

QuadRotorDynamics::QuadRotorDynamics(double k_t, double k_m, double moment_arm, int error_type, double K, double C, double T, double dead_time) :
    _k_t(k_t),
    _k_m(k_m),
    _moment_arm(moment_arm),
    _error_type(error_type),
    _K(K),
    _C(C),
    _T(T),
    _dead_time(dead_time),
    _rotor_speed_max(0.0),
    _rotor_acceleration_max(0.0),
    _limit_const_act(false),
    _limit_transient_act(false)
{}

// destructor
QuadRotorDynamics::~QuadRotorDynamics()
{}

Mechanics QuadRotorDynamics::update(ActuationVal actuation_value, double deltaT)
{
  // update actuation value
  _actuation_value = actuation_value;

  // update previous rotor speed 
  _rotor_speed_prev = _rotor_speed;

  // update delta T
  _dT = deltaT;

  // Quadrotor dynamics equations
  // pt1_element calculates the current rotor_speed
  QuadRotorDynamics::pt1_element();

  QuadRotorDynamics::check_limits();
 
  QuadRotorDynamics::aerodynamics();
  QuadRotorDynamics::roll_pitch_torque();
  
  QuadRotorDynamics::disturbance();

  // return updated values
  return _mechanics;

}

void QuadRotorDynamics::check_limits()
{
  if ( _limit_const_act && (_rotor_speed_max != 0.0) )
  { 
    // check limits of rotor speed
    if (_rotor_speed.motor[FRONT] > _rotor_speed_max)
    {
      std::cout << "DeadalusSim notification: Rotor speed reached its Limit: " << _rotor_speed_max << std::endl;
      _rotor_speed.motor[FRONT] = _rotor_speed_max;
      _rotor_speed.motor[REAR] = _rotor_speed_max;
      _rotor_speed.motor[LEFT] = _rotor_speed_max;
      _rotor_speed.motor[RIGHT] = _rotor_speed_max;
 
    }
  }
  
  if ( _limit_transient_act && (_rotor_acceleration_max != 0.0) )
  {
    // check limits of rotor acceleration 
    if (_rotor_acceleration.motor[FRONT] > _rotor_acceleration_max)
    {
      std::cout << "DeadalusSim notification: Rotor speed reached its Limit: " << _rotor_acceleration_max << std::endl;
      _rotor_speed.motor[FRONT] = _rotor_acceleration_max * _dT;
      _rotor_speed.motor[REAR] = _rotor_acceleration_max * _dT;
      _rotor_speed.motor[LEFT] = _rotor_acceleration_max * _dT;
      _rotor_speed.motor[RIGHT] = _rotor_acceleration_max * _dT;

    }
  }

}

void QuadRotorDynamics::pt1_element()
{
  if (_T != 0.0)
  {
    // K*u +C = T*d/dt*omega + omega
    _rotor_speed.motor[FRONT] = ( _actuation_value.motor[FRONT] * _K - _rotor_speed.motor[FRONT] + _C ) * _T / _dT + _rotor_speed.motor[FRONT];
    _rotor_speed.motor[REAR] = ( _actuation_value.motor[REAR] * _K - _rotor_speed.motor[REAR] + _C ) * _T / _dT + _rotor_speed.motor[REAR];
    _rotor_speed.motor[LEFT] = ( _actuation_value.motor[LEFT] * _K - _rotor_speed.motor[LEFT] + _C ) * _T / _dT + _rotor_speed.motor[LEFT];
    _rotor_speed.motor[RIGHT] = ( _actuation_value.motor[RIGHT] * _K - _rotor_speed.motor[RIGHT] + _C ) * _T / _dT + _rotor_speed.motor[RIGHT];

  }
  else
  {
    _rotor_speed.motor[FRONT] = _K*_actuation_value.motor[FRONT] + _C;
    _rotor_speed.motor[REAR] = _K*_actuation_value.motor[REAR] + _C;
    _rotor_speed.motor[LEFT] = _K*_actuation_value.motor[LEFT] + _C;
    _rotor_speed.motor[RIGHT] = _K*_actuation_value.motor[RIGHT] + _C;

  }

  for (int i = 0; i < 3; i ++)
  {
    if (_rotor_speed.motor[i] < 0)
    {
      _rotor_speed.motor[i] = 0.0;
    }
  }
  
  // deat time not supported yet
}

void QuadRotorDynamics::aerodynamics()
{
  // motor thrusts
  _mechanics.thrust.motor[FRONT] = _k_t * _rotor_speed.motor[FRONT] * _rotor_speed.motor[FRONT];
  _mechanics.thrust.motor[REAR] = _k_t * _rotor_speed.motor[REAR] * _rotor_speed.motor[REAR];
  _mechanics.thrust.motor[LEFT] = _k_t * _rotor_speed.motor[LEFT] * _rotor_speed.motor[LEFT];
  _mechanics.thrust.motor[RIGHT] = _k_t * _rotor_speed.motor[RIGHT] * _rotor_speed.motor[RIGHT];

  // yaw moment
  _mechanics.torque.yaw = _k_m * (_rotor_speed.motor[REAR] - _rotor_speed.motor[LEFT] + _rotor_speed.motor[FRONT] - _rotor_speed.motor[RIGHT]);
}

void QuadRotorDynamics::roll_pitch_torque()
{
  // calculate roll and pitch torques
  _mechanics.torque.roll = _moment_arm * _k_t * (_rotor_speed.motor[LEFT] - _rotor_speed.motor[RIGHT]); // TODO: Vorzeichen ueberpruefen
  _mechanics.torque.pitch = _moment_arm * _k_t * (_rotor_speed.motor[REAR] - _rotor_speed.motor[FRONT]); // TODO: Vorzeichen ueberpruefen

  if (_gyroscopic_torque_active == true)
  {
    //_mechanics.torque.roll = _mechanics.torque.roll + ...
    //_mechanics.torque.pitch = _mechanics.torque.pitch + ...
  }
}

void QuadRotorDynamics::disturbance()
{ 
  _sine_index += 1;
  switch (_error_type)
  {
    case NO_ERROR:
      //std::cout << "no_error" << std::endl;
    break;

    case GAUSSIAN_NOISE:
      //std::cout << "Gaussian noise not supported yet" << std::endl;
    break;
   
    case SINE_ERROR:
      //std::cout << "Sine error" << std::endl;
      _mechanics.thrust.motor[FRONT] = _mechanics.thrust.motor[FRONT];// + 2.0 + 2.0*cos(_sine_index);
      _mechanics.thrust.motor[REAR] = _mechanics.thrust.motor[REAR];// + 2.0 + 2.0*sin(_sine_index);
      _mechanics.thrust.motor[LEFT] = _mechanics.thrust.motor[LEFT] + 1.0*sin(_sine_index);
      _mechanics.thrust.motor[RIGHT] = _mechanics.thrust.motor[RIGHT] + 1.0*cos(_sine_index);
    break;
    
    default:
      std::cout << "Error type not supported." << std::endl;
  }
}

void QuadRotorDynamics::activate_limit_constant( bool active )
{
  _limit_const_act = active;
  if (active)
  {
    std::cout << "Constant limit is active now. " << std::endl;
    if (_rotor_speed_max == 0.0)
    {
      std::cout << "Maximal rotor speed is zero. Use set_limit_constant to set." << std::endl;
    }
   else
    { 
      std::cout << "Maximal value is " << _rotor_speed_max << "." << std::endl;
    }
  }
  else
  {
    std::cout << "Rotor speed max is deactivated." << std::endl;
  }

}

void QuadRotorDynamics::activate_limit_transient( bool active )
{
  _limit_transient_act = active;
  if (active)
  {
    std::cout << "Transient limit is active now. " << std::endl;
    if (_rotor_acceleration_max == 0.0)
    {
      std::cout << "Maximal rotor acceleration is zero. Use set_limit_constant() to set." << std::endl;
    }
    else
    {
      std::cout << "Maximal value is " << _rotor_acceleration_max << "." << std::endl;
    }
  }
  else
  {
    std::cout << "Rotor acceleration max is deactivated." << std::endl;
  }

}

void QuadRotorDynamics::activate_gyroscopic_torque( bool active )
{
  _gyroscopic_torque_active = active;
  if (_gyroscopic_torque_active)
  {
    std::cout << "Taking gyroscopic torque of the rotors into account is activated" << std::endl;
  }
  else
  {
    std::cout << "Taking gyroscopic torque of the rotors into account is deactivated" << std::endl;
  }
  
}

void QuadRotorDynamics::set_limit_const(double rotor_speed_max)
{
  _rotor_speed_max = rotor_speed_max;
}

void QuadRotorDynamics::set_limit_transient(double rotor_acceleration_max)
{
  _rotor_acceleration_max = rotor_acceleration_max;
}


void QuadRotorDynamics::set_k_pt1(double K)
{
  _K = K;
}
void QuadRotorDynamics::set_t_pt1(double T)
{
  _T = T;
}
void QuadRotorDynamics::set_c_pt1(double C)
{
  _C = C;
}
void QuadRotorDynamics::set_dead_time_pt1(double dead_time)
{
  _dead_time = dead_time;
}


Mechanics QuadRotorDynamics::getMechanics()
{
  return _mechanics;
}

Thrust QuadRotorDynamics::getThrust()
{
  return _mechanics.thrust;
}

Torque QuadRotorDynamics::getTorque()
{
  return _mechanics.torque;
}

void QuadRotorDynamics::setErrorType(int error_type)
{
  _error_type = error_type;
}
