/* -----------------------------------------------------------------------------
 *
 *
 *  ____                   _____     _
 * / ___|  __ _ _ __  _ __|__  /  __| | ___
 * \___ \ / _` | '_ \| '_ \ / /  / _` |/ _ \
 *  ___) | (_| | |_) | |_) / /_ | (_| |  __/
 * |____/ \__,_| .__/| .__/____(_)__,_|\___|
 *             |_|   |_|
 *
 *
 *
 * Author: Andreas Gschossmann
 * Email:  andreas.gschossmann@hs-regensburg.de
 *
 * quadrotor_dynamics.hh - Relationship between rotor-speed or alternatively
 * digital value for brushless drivers and resulting thrust of the motor.  *
 * -----------------------------------------------------------------------------*/
#ifndef _GAZEBO_DYNAMICS_HH_ 
#define _GAZEBO_DYNAMICS_HH_

#include <string>

#include <math.h>
#include <gazebo/math/Vector3.hh>

#include "../dependencies/resources.hh"
#include "../dependencies/constants.hh"

/// \brief  Class for the calculation of the Quadrotor Dynamics 
///
/// This class calculates the Torques and Forces on the quadrotor.
/// Since gazebo does not support aerodynamics, this class calculates
/// the Thrust Forces and the Yaw Torques produced by the the rotor 
/// speeds of the motors, which is resulting from the current actuation
/// value.

class QuadRotorDynamics 
{ 
    /// \brief Constructor.
    ///
    /// constructor with default values
    public: QuadRotorDynamics(
                 double k_t = 1.0,		///< thrust coefficient
                 double k_m = 1.0, 		///< moment coefficient 
                 double moment_arm = 0.220,	///< distance between motor axis and center of frame 
                 int error_type = NO_ERROR,	///< error type for disturbance function() 
                 double K = 1.0, 		///< K-constant for pt1_element() 
                 double C = 0.0, 		///< C-constant for pt1_element()
                 double T = 0.0, 		///< T-constant for pt1_element()
                 double dead_time = 0.0		///< dead_time for pt1_element() (not supported yet)
                 ); 

    /// \brief Destructor.
    public: virtual ~QuadRotorDynamics();

    /// \brief Updates the Thrusts and Torques 
    ///
    /// This function updates the values of the Thrusts and Torques
    /// resulting from the current rotor speeds. It is intended to
    /// be called in every simulation step.
    /// @param actuation_value 	actuation value corresponding to desired rotor speed
    /// @param deltaT 		simulatin time passed since the last simulation step
    /// @return			The Thrust and Torque on the quadrotor frame 
    public: Mechanics update(ActuationVal actuation_value, double deltaT);
    //public: Mechanics update(AngularVel rotor_speed, double deltaT);
  
    /// \brief Check limits of maximal rotor speed and maximal rotor acceleration
    /// 
    /// If activated by activate_limit_constant(bool) and if the maximal rotor speed
    /// (_rotor_speed_max) the maximal rotor acceleration is limited to the maximal
    /// value, set by set_limit_const.
    /// If activated by activate_limit_transient(bool) and if the maximal rotor acceleration
    /// value, set by set_limit_transient.
    private: void check_limits();
    
    /// \brief Calculates the Thrust Forces and Drag Torque on the quadrotor frame created
    /// by the rotors angular velocities
    private: void aerodynamics();

    /// \brief Calculates the roll and pitch torques on the quadrocopter frame.
    ///
    /// This function calculates the roll and pitch torques, resulting the thrust
    /// forces, calculated by aerodynamics. If activated by activate_gyroscopic_torque(bool)
    /// also the torques resulting from the gyroscopic effect are being considered
    private: void roll_pitch_torque();

    /// \brief Provides disturbance to the Thrust
    ///
    /// Provides one of the supported kinds of disturbance to the Thrust, calculated 
    /// by aerodynamics(). The supported error-types are NO_ERROR, GAUSSIAN_NOISE,
    /// SINE_ERROR. They are set by setErrorType(int error_type) and stored to the
    /// private member _error_type. 
    private: void disturbance();

    private:
      AngularVel     	_rotor_speed; 		///< Current rotor speed
      AngularVel	_rotor_speed_prev;	///< Previous rotor speed
      AngularVel     	_rotor_acceleration;	///< Current rotor acceleration
      ActuationVal 	_actuation_value;	///< Current actuation value 
      Mechanics 	_mechanics;		///< Current Mechanics, containing Thrust Forces and Drag Torques
      double    _dT;				///< Simulation time, which has passed since the last simulation step

    // aerodynamics constants
    private:
      double _k_t;		///< Thrust coefficient 
      double _k_m;		///< Moment coefficient
      
    private:
      double _moment_arm;      	///< Distance between rotor axis and center of quadrotor frame

    private:
      int _error_type;  	///< Error for the error provided by disturbance and set by setErrorType(int error_type) 
      double _sine_index;	///< Index needed by disturbance for the error type SINE_ERROR
  
    // For PT1 K*u + C = T*(d/dt)omega + omega 
    private:
      double _K;		///< K coefficient for pt1_element()
      double _C;		///< C coefficient for pt1_element()
      double _T;		///< T coefficient for pt1_element()
      double _dead_time; 	///< Dead Time for pt1_elment() (not supported yet)
    
    /// \brief PT1 element K*u + C = T*(d/dt)omega + omega 
    private: void pt1_element(); // K*u + C = T*(d/dt)omega + omega
   
    // values for limit check 
    private:
      double _rotor_speed_max; 		///< Maximal rotor speed limited by check_limits()
      double _rotor_acceleration_max;	///< Maximal rotor accaleration limited by check_limits()
      bool   _limit_const_act;		///< If set true, maximal rotor speed limit is checked by 
					///< check_limits(), it is set by activate_limit_constant( bool activate )
      bool   _limit_transient_act;	///< If set true, maximal rotor acceleration limit is checked by
                                        ///< check_limits(), it is set by activate_limit_transient( bool activate )

    private:
      bool   _gyroscopic_torque_active;	///< If set true, the gyroscopic effect is considered by roll_pitch_torque()
                                        ///< it is set by activate_gyroscopic_torque( bool activate )

    // for check_limits()
    public: void activate_limit_constant( bool active ); 		///< Makes check_limits() limit the maximal rotor speed
    public: void activate_limit_transient( bool active );		///< Makes check_limits() limit the maximal rotor acceleration
    public: void set_limit_const(double rotor_speed_max);		///< Set maximal rotor speed for check_limits() 
    public: void set_limit_transient(double rotor_acceleration_max);	///< Set maximal rotor acceleration for check_limits()

    public: void activate_gyroscopic_torque( bool active );		///< Activates/Deactivates consideration of gyroscopic effect in roll_pitch_torque()
   
    // set functions for pt1 element
    public: void set_k_pt1(double K); 			///< Set K coefficient for pt1_element()
    public: void set_t_pt1(double T);			///< Set T coefficient for pt1_element()
    public: void set_c_pt1(double C);			///< Set C coefficient for pt1_elmennt()
    public: void set_dead_time_pt1(double dead_time);	///< Set dead time for pt1_element() (not supported yet)
     
    public:
      Mechanics getMechanics();	///< Returns the Mechanics on the quadrotor frame including Thrust Forces and Drag Torques
      Thrust    getThrust();	///< Returns Thrust Forces produced by the motors 
      Torque    getTorque();	///< Returns Torque on the quadrotor frame

    public: void setErrorType(int error_type); ///< Sets the error type needed by disturbance() (NO_ERROR, GAUSSIAN_NOISE and SINE_ERROR supported)

};
#endif
