/* -----------------------------------------------------------------------------
 *
 *
 *  ____                   _____     _
 * / ___|  __ _ _ __  _ __|__  /  __| | ___
 * \___ \ / _` | '_ \| '_ \ / /  / _` |/ _ \
 *  ___) | (_| | |_) | |_) / /_ | (_| |  __/
 * |____/ \__,_| .__/| .__/____(_)__,_|\___|
 *             |_|   |_|
 *
 *
 *
 * Author: Andreas Gschossmann
 * Email:  andreas.gschossmann@hs-regensburg.de
 *
 * quadrotor_dynamics.hh - Relationship between rotor-speed or alternatively
 * digital value for brushless drivers and resulting thrust of the motor.
 *
 * -----------------------------------------------------------------------------*/
#ifndef _SIL_HH_ 
#define _SIL_HH_

#include "../dependencies/resources.hh"
#include "../dependencies/constants.hh"
#include "../joystick/joystick.hh"

/// \brief The SIL class contains everything, that is needed to simulate the
/// Software in the Loop (SIL) of the quadrocopter, which is intended to be
/// the quadrotor firmware

class SIL 
{ 
    /// \brief The SIL class contains everything, that is needed to simulate
    /// the firmware of the quadrocopter 
    public: SIL();

    /// \brief Destructor.
    public: virtual ~SIL();
  
    private: 
      AngularVel 	_rotor_speed;			///< Current rotor speed
      ActuationVal	_actuation_value;		///< Current actuation value

    private:
      int 		_test_sequence_iterator;	///< Iterator for the test_sequence function

    private:
      ///////////////////////////////////
      // variables for C-functions in main loop
      ///\cond
      float	acc_x, acc_y, acc_z;	 ///\internal data accelerometer
      float	gyro_x, gyro_y, gyro_z;  // data gyro
      float	gx, gy, gz;
      float 	roll, pitch, yaw;	 // inertial angles
      float	y_roll, y_pitch, y_yaw;
      float	gas_ctrl, pitch_ctrl, roll_ctrl, yaw_ctrl; // desired values from controller 
      ///\endcond
      //////////////////////////////////

    private:
       Joystick _ps3controller;			///< Instance of Joystick class to obtain data of playstion 3 controller

    private: void test_sequence();		///< This function contains a sequence of test stimuli. It is intended to be called in update(double, Sensors)

    private: void get_ps3controller();		///< Obtains the current state of the playstation 3 controller
    private: void get_key();			///< Obtains keyboard inputs (not supported yet) 
    
    public: 
      AngularVel	get_rotor_speed();	///< Returns the current rotor speed
      ActuationVal	get_actuation_value(); 	///< Returns the current actuation value
   
     
    public: void init();					///< \brief This function is intended to be called when the 
   								///< simulation is loaded. It should contain all test code
                                                                ///< which is intended to be called before the main loop

    public: void update(double dT, Sensors mySensorsVals); 	///< This function is intended to be called in the
      								///< whenever the simulation is updated. It should 
                                                                ///< contain the code of the firmware which is intended
                                                                ///< to be simulated.

  };

#endif
